app.service('BaseChart', function ($log) {
	var debug = false;
	return {
		createChart: function(scope, element) {
			var self = this;
			
			// Create the chart object
			var chart = {
				id:			scope.elementId,
				scope:		scope,
				element: 	element,
				options:	scope.options,
				disabled:	{},
				height:		0,
				width:		0
			};
			
			// Create lists of functions
			chart.draws = [];
			chart.updates = [];
			
			// Register draw and update functions
			chart.register = function(type, func) {
				if(type == 'draw')
					chart.draws.push(func);
				else if(type == 'update')
					chart.updates.push(func);
			};
			
			// Clear the existing chart
			chart.clear = function() {
				d3.select(chart.element[0]).select('svg').selectAll("*").remove();
			};
			
			// Draw the chart
			chart.draw = function() {
				// Clear existing elements
				chart.clear();
				
				// Get the overall chart height and width from the parent
				chart.height = $(chart.element).parent().height();
				chart.width = $(chart.element).parent().width();

				// Draw the chart canvas
				self.drawCanvas(chart);
			
				// If there is a title, draw this now
				if(chart.options.chartTitle != undefined)
					self.drawTitle(chart);

				// If there is a legend, draw this now
				if(chart.options.legend == true && chart.options.key != undefined)
					self.drawLegend(chart);
				
				// Draw the plot area
				self.drawPlotArea(chart);
					
				// If there are draw functions, call these now
				for(var idx = 0; idx < chart.draws.length; idx++)
					chart.draws[idx](this);
				
				// Call chart update
				var isAnimated = true;
				chart.update(isAnimated);
			};
			
			// Update the chart
			chart.update = function(isAnimated) {
				// If there is a legend, update this and the plot area now
				if(chart.options.legend == true && chart.options.key != undefined) {
					self.updateLegend(chart);
					self.updatePlotArea(chart);
				}

				// If there are update functions, call these now
				for(var idx = 0; idx < chart.updates.length; idx++)
					chart.updates[idx](this, isAnimated);				
			};
			
			// Tooltip function
			chart.tooltip = function(action, tooltip) {
				self.tooltip(this, action, tooltip);
			};
			
			// Disable a given data point
			chart.disable = function(d) {
				chart.disabled[d[chart.options.uid]] = d;
			};
			
			// Enable a given data point
			chart.enable = function(d) {
				delete chart.disabled[d[chart.options.uid]];
			};
			
			// Is a given data point enabled?
			chart.isDisabled = function(d) {
				if(chart.disabled[d[chart.options.uid]] != undefined)
					return true;
				return false;
			};
			

			return chart;
		},

		drawCanvas: function(chart) {
			// Create the SVG element (RED)
			chart.canvas = {
				width:	chart.width,
				height:	chart.height
			};

			var svg = d3.select(chart.element[0])
				.select('svg')
					.attr('width', chart.width)
					.attr('height', chart.height)
					.attr('class', 'chart-svg');
			chart.canvas.area = svg;
			if(debug == true) svg.append('rect').attr('width', chart.width).attr('height', chart.height).style('fill', 'red');
			
			// Create the chart area - Canvas minus margins (BLUE)
			chart.chartArea = {
				width:	chart.width - chart.options.margin.left - chart.options.margin.right,
				height:	chart.height - chart.options.margin.top - chart.options.margin.bottom,
				x:		chart.options.margin.left,
				y:		chart.options.margin.top
			};
			var chartArea = svg
				.append('g')
					.attr('width', chart.chartArea.width)
					.attr('height', chart.chartArea.height)
					.attr('class', 'chart-area')
					.attr('transform', 'translate(' + chart.chartArea.x +  ',' + chart.chartArea.y + ')');
			chart.chartArea.area = chartArea;
			if(debug == true) chartArea.append('rect').attr('width', Math.max(chart.chartArea.width, 0)).attr('height',  Math.max(chart.chartArea.height, 0)).style('fill', 'blue');

			// Create the detail area - Detail area to start with (PINK)
			chart.detailArea = {
				width:	chart.chartArea.width,
				height:	chart.chartArea.height,
				x:		chart.options.margin.left,
				y:		chart.options.margin.top
			};
			var detailArea = chartArea
				.append('g')
					.attr('width', chart.detailArea.width)
					.attr('height', chart.detailArea.height)
					.attr('class', 'detail-area');	
			chart.detailArea.area = detailArea;
			if(debug == true) chart.detailArea.rect = detailArea.append('rect').attr('width', Math.max(chart.detailArea.width, 0)).attr('height',  Math.max(chart.detailArea.height, 0)).style('fill', 'pink');					
		},
		
		drawTitle: function(chart) {
			// Setup configuration
			var titlePadding = 10;
			
			// Setup areas
			var chartArea = chart.chartArea.area;
			var detailArea = chart.detailArea.area;
			
			// Create title area (ORANGE)
			chart.titleArea = {
				width:	chart.chartArea.width,
				height:	chart.chartArea.height,
				x:		0,
				y:		0
			};
			var titleArea = chartArea
				.append('g')
					.attr('width', chart.titleArea.width)
					.attr('height', chart.titleArea.height)
					.attr('class', 'title-area');
			chart.titleArea.area = titleArea;
			if(debug == true) chart.titleArea.rect = titleArea.append('rect').attr('width', Math.max(chart.titleArea.width, 0)).attr('height',  Math.max(chart.titleArea.height, 0)).style('fill', 'orange');

			// Create the title
			var title = titleArea
				.append('text')
					.attr('x', 0)
					.attr('y', 0)
					.attr('class', 'chart-title')
					.text(chart.options.chartTitle);
			chart.title = {
				width:	title[0][0].getBBox().width,
				height:	title[0][0].getBBox().height
			};
			
			// Update the title area height
			chart.titleArea.height = chart.title.height + titlePadding;
			titleArea.attr('height', chart.titleArea.height);
			if(debug == true) chart.titleArea.rect.attr('height', Math.max(chart.titleArea.height, 0));

			// Calculate size and position
			var titleX = (chart.titleArea.width - chart.title.width) / 2;
			var titleY = chart.title.height;
			
			// If a title offset is configured, calculate that
			if(chart.options.chartTitleOffset != undefined) {
				if(chart.options.chartTitleOffset.x != undefined)
					titleX = titleX + chart.options.chartTitleOffset.x;
				if(chart.options.chartTitleOffset.y != undefined)
					titleY = titleY + chart.options.chartTitleOffset.y;
			}
			
			// Apply transformation for positioning
			title.attr('transform', 'translate(' + titleX +  ',' + titleY + ')');

			// Update the detail area height and position
			chart.detailArea.height = chart.detailArea.height - chart.titleArea.height;					
			detailArea
				.attr('height', chart.detailArea.height)
				.attr('transform', 'translate(0,' + chart.titleArea.height + ')');				
			if(debug == true) chart.detailArea.rect.attr('height', Math.max(chart.detailArea.height, 0));							
		},

		drawLegend: function(chart) {
			// Setup areas
			var detailArea = chart.detailArea.area;

			// Create legend area (GREY)
			chart.legendArea = {
				width:	chart.detailArea.width,
				height:	chart.detailArea.height,
				x:		0,
				y:		0
			};
			var legendArea = detailArea
				.append('g')
					.attr('width', Math.max(chart.legendArea.width, 0))
					.attr('height', Math.max(chart.legendArea.height, 0))
					.attr('class', 'legend-area');
			chart.legendArea.area = legendArea;
			if(debug == true) chart.legendArea.rect = legendArea.append('rect').attr('width', Math.max(chart.legendArea.width, 0)).attr('height',  Math.max(chart.legendArea.height, 0)).style('fill', 'grey');
		},
		
		updateLegend: function(chart) {
			var self = this;

			// Setup areas
			var legendArea = chart.legendArea.area;
			
			// Setup configuration
			var legendRectSize	= chart.options.legendSize;
			var legendSpacing	= chart.options.legendSpacing;
			var columnSpacing 	= legendRectSize + legendSpacing;
			
			// Reset the legend area dimensions
			chart.legendArea.width = chart.detailArea.width;
			chart.legendArea.height = chart.detailArea.height;

			// Generate the legend dataset
			var values = d3.values(chart.scope.dataset)
				.sort(function(a, b){ return self.sort(chart, a, b); });
				
			var legend = legendArea.selectAll('.chart-legend')
				.data(values, function(d){ return d[chart.options.uid]; });

			// Delete removed nodes
			legend.exit()
				.each(function(d) {	d3.select(this).selectAll('*').remove(); })
				.remove();
			
			// Insert added nodes
			legend.enter()
				.append('g')
					.attr('class', 'chart-legend')
				.each(function(d) {
					d3.select(this).append('rect')
						.attr('class', function(d, i) {	return chart.isDisabled(true) ? 'chart-legend-rectangle disabled' : 'chart-legend-rectangle'; })
						.attr('width', legendRectSize)
						.attr('height', legendRectSize)
						.style('fill', function(d, i) { return chart.options.color(d[chart.options.colorValue]); })
						.style('stroke', function(d, i) { return chart.options.color(d[chart.options.colorValue]); })
						.on('click', function(d, i) {
							var node = d3.select(this.parentNode);
							var rect = d3.select(this);
							var total = legendArea.selectAll('.chart-legend')[0].length;
							var count = legendArea.selectAll('.disabled')[0].length;
							var disabled = chart.isDisabled(d);
							$log.debug(disabled);
							if(disabled == true) {
								rect.attr('class', rect.attr('class').replace('disabled').trim());
								chart.enable(d);
							}
							else if(count + 1 < total) {
								rect.attr('class', rect.attr('class') + ' disabled');
								chart.disable(d);
							}			
							var isAnimated = true;
							chart.update(isAnimated);
						});  
					d3.select(this).append('text');
				});
			
			// Update each legend element
			var maxLegendTitleWidth = 0;
			var maxLegendTitleHeight = 0;
			legend.select('text')
				.text(function(d) { return d[chart.options.key]; })
				.attr('x', function(d) { return legendRectSize + legendSpacing; })
				.attr('y', function(d) { return legendRectSize / 2 + this.getBBox().height / 4; })
				.attr('class', function(d) {
					var c = 'chart-legend-label';
					if(chart.options.selectable == false) return c;	
					var selected = (chart.scope.selected != undefined && chart.scope.selected[chart.options.uid] == d[chart.options.uid]);
					if(selected == true)
						return c + ' selected';
					return c;
				})
				.each(function(d) {
					maxLegendTitleWidth = Math.max(maxLegendTitleWidth, this.getBBox().width + legendRectSize + legendSpacing);
					maxLegendTitleHeight = Math.max(maxLegendTitleHeight, Math.max(this.getBBox().height + legendSpacing, legendRectSize + legendSpacing));
				});
		
			// Position each element					
			if(chart.options.legendPosition == "left" || chart.options.legendPosition == "right") {
				var col = 1;
				var row = 1;
				var offset = 0;
				legend.attr('transform', function(d, i) {
					var horz = (col - 1) * (maxLegendTitleWidth + columnSpacing) + legendSpacing;
					var vert = i * (legendRectSize + legendSpacing) - (col - 1) * offset + legendSpacing;
					row++;
					if(vert + maxLegendTitleHeight > chart.legendArea.height) {
						col++;
						row = 1;
						if(offset == 0)	offset = vert - legendSpacing;
						var horz = (col - 1) * (maxLegendTitleWidth + columnSpacing) + legendSpacing;
						vert = vert - offset;
					}							
					return 'translate(' + horz + ',' + vert + ')';
				});		
				
				// Update legendArea dimensions and position
				chart.legendArea.height = offset > 0 ? offset + legendSpacing : (row - 1) * maxLegendTitleHeight + legendSpacing;
				chart.legendArea.width = col * (maxLegendTitleWidth + columnSpacing) - columnSpacing + 2 * legendSpacing;
				if(chart.options.legendPosition == "right")
					chart.legendArea.x = chart.detailArea.width - chart.legendArea.width;
				chart.legendArea.y = (chart.detailArea.height - chart.legendArea.height) / 2;
				
				// Apply dimensions and position
				legendArea
					.attr('height', Math.max(chart.legendArea.height, 0))
					.attr('width', Math.max(chart.legendArea.width, 0))
					.attr('transform', 'translate(' + chart.legendArea.x +  ',' + chart.legendArea.y + ')');
				if(debug == true) chart.legendArea.rect.attr('height', chart.legendArea.height).attr('width', chart.legendArea.width);
			}
			
			if(chart.options.legendPosition == "top" || chart.options.legendPosition == "bottom") {
				var hOffset = 0;
				var vOffset = 0;
				var maxWidth = 0;
				legend.attr('transform', function(d, i) {
					var w = this.getBBox().width + legendSpacing;							
					var horz = hOffset + legendSpacing;
					var vert = vOffset + legendSpacing;
					hOffset = hOffset + w;
					maxWidth = Math.max(maxWidth, hOffset);
					if(horz + w > chart.legendArea.width) {
						hOffset = 0;
						vOffset = vOffset + maxLegendTitleHeight + legendSpacing;
						horz = hOffset + legendSpacing;
						vert = vOffset + legendSpacing;
						hOffset = hOffset + w;
					}
					
					return 'translate(' + horz + ',' + vert + ')';
				});		
				
				// Update legendArea dimensions and position
				chart.legendArea.height = vOffset + maxLegendTitleHeight + legendSpacing;
				if(chart.options.legendPosition == "bottom")
					chart.legendArea.y = chart.detailArea.height - chart.legendArea.height;
				legendArea
					.attr('height', Math.max(chart.legendArea.height, 0))
					.attr('width', Math.max(chart.legendArea.width, 0))
					.attr('transform', 'translate(' + chart.legendArea.x +  ',' + chart.legendArea.y + ')');							
				if(debug == true) chart.legendArea.rect.attr('height', chart.legendArea.height).attr('width', chart.legendArea.width);
			}
			
		},
		
		drawPlotArea: function(chart) {
			var plotAreaPadding = 10;
			
			// Setup areas
			var detailArea = chart.detailArea.area;

			// Create the plot area (YELLOW)
			chart.plotArea = {
				width:			chart.detailArea.width,
				height:			chart.detailArea.height,
				padding:		plotAreaPadding,
				paddingDefault:	plotAreaPadding,
				x:				0,
				y:				0
			};
			var plotArea = detailArea
				.append('svg')
					.attr('width', Math.max(chart.plotArea.width, 0))
					.attr('height', Math.max(chart.plotArea.height, 0))
					.attr('class', 'plot-area')
					.attr('x', chart.plotArea.x)
					.attr('y', chart.plotArea.y);
			chart.plotArea.area = plotArea;
			if(debug == true) chart.plotArea.rect = plotArea.append('rect').attr('width', Math.max(chart.plotArea.width, 0)).attr('height',  Math.max(chart.plotArea.height, 0)).style('fill', 'yellow');
		},
		
		updatePlotArea: function(chart) {
			// Setup areas
			var plotArea = chart.plotArea.area;
			
			// Adjust the padding
			var p = chart.detailArea.width - chart.legendArea.width - chart.plotArea.width;
			if(p > 0)
				chart.plotArea.padding = p;
			else
				chart.plotArea.padding = chart.plotArea.paddingDefault;
			
			// Reset the plot area size to defaults
			chart.plotArea.width = chart.detailArea.width;
			chart.plotArea.height = chart.plotArea.height;
			
			// Modify plot area if there is a legend
			if(chart.options.legend == true && chart.options.key != undefined) {
				if(chart.options.legendPosition == "left") {
					chart.plotArea.width = chart.plotArea.width - chart.legendArea.width - chart.plotArea.padding;
					chart.plotArea.x = chart.legendArea.width  + chart.plotArea.padding;
				}
				else if(chart.options.legendPosition == "right")
					chart.plotArea.width = chart.plotArea.width - chart.legendArea.width - chart.plotArea.padding;
				else if(chart.options.legendPosition == "top") {
					chart.plotArea.height = chart.plotArea.height - chart.legendArea.height;
					chart.plotArea.y = chart.legendArea.height;
				}
				else if(chart.options.legendPosition == "bottom")
					chart.plotArea.height = chart.plotArea.height - chart.legendArea.height;
			}

			// Change size and position
			plotArea
				.attr('width', Math.max(chart.plotArea.width, 0))
				.attr('height', Math.max(chart.plotArea.height, 0))
				.attr('x', chart.plotArea.x)
				.attr('y', chart.plotArea.y);
			if(debug == true) chart.plotArea.rect.attr('width', Math.max(chart.plotArea.width, 0)).attr('height',  Math.max(chart.plotArea.height, 0));			
		},
		
		sort: function(chart, a, b) {
			var self = this;
			
			var val = 0;
			if(chart.options.groupBy == true)
				val = self.sortByGroup(chart, a, b);
				
			if(val != 0) return val;
			return self.sortByValue(chart, a, b);
		},
		
		sortByGroup: function(chart, a, b) {
			var val = 0;
			if (a[chart.options.groupByValue] < b[chart.options.groupByValue])
				val = -1;
			else if (a[chart.options.groupByValue] > b[chart.options.groupByValue])
				val = 1;						
			var dir = chart.options.sortDirection == "desc" ? -1 : 1;
			val = dir * val;		
			return val;
		},
		
		sortByValue: function(chart, a, b) {
			var val = 0;
			if(chart.options.sorted == true) {
				if (a[chart.options.sortValue] < b[chart.options.sortValue])
					val = -1;
				else if (a[chart.options.sortValue] > b[chart.options.sortValue])
					val = 1;		
				var dir = chart.options.sortDirection == "desc" ? -1 : 1;
				val = dir * val;		
			}
			return val;
		},
		
		tooltip: function(chart, action, tooltip) {
			var tooltipTransition = 300;
			var container = $(chart.element).parent();
			var div = d3.select(chart.element[0]).select(".chart-tooltip");
			if(action == 'show') {
				var width = $(chart.element[0]).find('.chart-tooltip').width();
				var height = $(chart.element[0]).find('.chart-tooltip').height();

				var x = tooltip.position[0] - width / 2;
				var y = tooltip.position[1] - height / 2;				
				if(tooltip.position[0] + width + container.offset().left + 20 >= window.innerWidth)
					x = tooltip.position[0] - width - container.offset().left - 20;
				
				div.style("left", x + "px")     
					.style("top", y + "px")
					.style("display", "block")
					.style('opacity', 0.9);
				chart.scope.tooltipData = tooltip;
			}
			else if(action == 'hide') {
				div.style("display", "none")
					.style('opacity', 0);
				delete chart.scope.tooltipData;
			}
		},
	};
});

app.service('PieChart', function ($log, BaseChart) {
	var debug = false;
	return {
		createChart: function(scope, element) {
			var self = this;

			var chart = BaseChart.createChart(scope, element);
			chart.register('draw', self.draw);
			chart.register('update', self.update);
			
			// Setup the pie chart
			chart.pie = {
				outerRadius: 		0,
				innerRadius:		0,
				growOnHoverRadius:	0,
				selectInnerRadius:	0,
				selectOuterRadius:	0,
				labelRadius:		0,
				startAngle:			0,
				endAngle:			360
			};
			
			// Setup quadrants
			var quadrants = {
				"N":	[90,-90],
				"S":	[90,270],
				"E":	[0,180],
				"W":	[180,360],
				"NE":	[0,90],
				"NW":	[0,-90],
				"SE":	[90,180],
				"SW":	[180,270]
			};
			
			// If chart is quadrant, then apply to configuration
			if(chart.options.quadrant != undefined && chart.options.quadrant != "") {
				chart.pie.startAngle = quadrants[chart.options.quadrant][0];
				chart.pie.endAngle = quadrants[chart.options.quadrant][1];
			}
			
			// Create a calculate radii function
			chart.calculateRadii = function() {
				self.calculateRadii(chart);
			};
			
			// Create an add listeners function
			chart.addListeners = function() {
				self.addListeners(chart);
			};
			
			// Watch for changes in hover and apply the grow on hover if enabled
			chart.scope.$watch('hover', function(newVal, oldVal) {
				if(newVal != undefined) {
					d3.select(newVal.path).attr('class', function(d) { return d3.select(this).attr('class') + ' hover' });
					if(chart.options.growOnHover == true) {
						d3.select(newVal.path).transition()
							.duration(70)
							.attr("d", chart.growArc);
					}
					var angle = (chart.pie.endAngle < 0 ? 360 + chart.pie.endAngle : chart.pie.endAngle) - chart.pie.startAngle;
					var pct = Math.abs((newVal.d.endAngle - newVal.d.startAngle) / (angle * Math.PI / 180));
					var tooltip = {
						color: 		chart.options.color(newVal.d.data[chart.options.colorValue]),
						payload:	newVal.d,
						percentage:	pct,
						position:	d3.mouse(chart.canvas.area[0][0])
					};
					chart.tooltip('show', tooltip);					
				}
				if(oldVal != undefined) {
					d3.select(oldVal.path).attr('class', function(d) {  return d3.select(this).attr('class').replace('hover', '').trim() });					
					if(chart.options.growOnHover == true) {
						d3.select(oldVal.path).transition()
							.duration(70)
							.attr("d", chart.pieArc);				
					}
					chart.tooltip('hide');					
				}
			});

			// Display the context menu
			$(element).on("contextmenu", function(e) {
				var container = $(element).parent();
				var menu = $(element).find(".context-menu");
				
				var x = e.pageX - container.offset().left;
				var y = e.pageY - container.offset().top;
				var width = menu.width();
				
				if(e.pageX + width > window.innerWidth)
					x = e.pageX - width - container.offset().left;
				
				$('.context-menu').hide();
				menu.css({
					display: "block",
					left: x,
					top: y
				});

				return false;
			});			
			
			// Display the query details modal
			scope.displayQuery = function() {
				scope.modal.mode = "display";
				scope.modal.query = scope.query;
				scope.modal.onSuccess = function() {
					// If the table is in a tab, select the tab this table is in on successful
					//   complete of the modal actions
					var tabs = $(element).closest('div.tab-content').prev('ul.nav-tabs');
					if(tabs.length > 0) {
						var showId = tabs.find('.active').find('a[data-toggle="tab"]').attr('href');
						setTimeout(function() {
							var showTab = tabs.find('a[href=' + showId + ']');
							showTab.tab('show');
							showTab.trigger("shown.bs.tab");
						}, 10);
					}
				};
				$(element).closest('.application').find('query-modal').find('.modal').modal('show');
			};
						
			return chart;
		},
		
		draw: function(chart) {
			// Setup areas
			var plotArea = chart.plotArea.area;
			
			// Create a pie area (PURPLE)
			chart.pieArea = {
				width:	chart.plotArea.width,
				height:	chart.plotArea.height,
				x:		0,
				y:		0
			};
			var pieArea = plotArea
				.append('g')
					.attr('width', chart.pieArea.width)
					.attr('height', chart.pieArea.height)
					.attr('class', 'chart-pie')
					.attr('transform', 'translate(' + chart.pieArea.x +  ',' + chart.pieArea.y + ')');
			chart.pieArea.area = pieArea;
			if(debug == true) chart.pieArea.rect = pieArea.append('rect').attr('width', Math.max(chart.pieArea.width, 0)).attr('height',  Math.max(chart.pieArea.height, 0)).style('fill', 'purple');
			
		},

		update: function(chart, isAnimated) {
			// Disable hover actions
			chart.hoverEnabled = false;

			// Setup areas
			var pieArea = chart.pieArea.area;
			
			// Reset the pie area dimensions to default
			chart.pieArea.width = chart.plotArea.width;
			chart.pieArea.height = chart.plotArea.height;
			
			// Calculate the radii for chart elements
			chart.calculateRadii(chart);

			// Define the pie arcs
			chart.pieArc = d3.svg.arc()
				.innerRadius(chart.pie.innerRadius)
				.outerRadius(chart.pie.outerRadius);
	
			chart.growArc = d3.svg.arc()
				.innerRadius(chart.pie.innerRadius)
				.outerRadius(chart.pie.growOnHoverRadius);
			
			chart.selectArc = d3.svg.arc()
				.innerRadius(chart.pie.selectInnerRadius)
				.outerRadius(chart.pie.selectOuterRadius);
				
			chart.groupArc = d3.svg.arc()
				.innerRadius(chart.pie.groupInnerRadius)
				.outerRadius(chart.pie.groupOuterRadius);			

			// Define the pie layout
			chart.pie.layout = d3.layout.pie()
				.value(function(d) { return chart.isDisabled(d) == true ? 0 : d[chart.options.value]; })
				.sort(function(a, b){ return BaseChart.sort(chart, a, b); })
				.startAngle(chart.pie.startAngle * (Math.PI/180))
				.endAngle(chart.pie.endAngle * (Math.PI/180));

			// Define pie coordinates
			chart.pie.x = chart.pieArea.width / 2;
			chart.pie.y = chart.pieArea.height / 2;
			if(chart.options.quadrant == "W" || chart.options.quadrant == "NW" || chart.options.quadrant == "SW")
				chart.pie.x = chart.pie.x + chart.pie.outerRadius / 2;
			if(chart.options.quadrant == "E" || chart.options.quadrant == "NE" || chart.options.quadrant == "SE")
				chart.pie.x = chart.pie.x - chart.pie.outerRadius / 2;
			if(chart.options.quadrant == "N" || chart.options.quadrant == "NE" || chart.options.quadrant == "NW")
				chart.pie.y = chart.pie.y + chart.pie.outerRadius / 2;
			if(chart.options.quadrant == "S" || chart.options.quadrant == "SE" || chart.options.quadrant == "SW")
				chart.pie.y = chart.pie.y - chart.pie.outerRadius / 2;
			
			// Generate the pie dataset
			var pie = pieArea.selectAll('.chart-pie-arc')
				.data(chart.pie.layout(d3.values(chart.scope.dataset)), function(d){ return d.data[chart.options.uid]; });
			
				// Delete removed nodes
				pie.exit().remove();
			
			// Insert added nodes
			pie.enter()
				.append('g')
					.attr('class', 'chart-pie-arc')
				.each(function(d) {
					d3.select(this).append('path')
						.style('fill', function(d) { return chart.options.color(d.data[chart.options.colorValue]); })
						.attr('class', function(d) { return 'chart-pie-slice' + (chart.options.selectable == true ? ' selectable' : ''); })
						.each( function(d) { this._current = d; })						
						.on('mouseover', function(d) { chart.scope.hover = {path: this, d: d}; chart.scope.$apply(); })
						.on('mouseout', function(d) { delete chart.scope.hover; chart.scope.$apply(); })
						.on('click', function(d) {				
							if(chart.options.selectable == false) return;	

							// Is the clicked pie slice already selected, if so set the selection
							//   otherwise remove the selection
							var selected = (chart.scope.selected != undefined && chart.scope.selected[chart.options.uid] == d.data[chart.options.uid]);
							if(selected == false)
								chart.scope.selected = d.data;
							else
								delete chart.scope.selected;

							// Hide the tooltip if there is one
							chart.tooltip('hide');

							// Update the chart
							chart.scope.$apply();
							chart.update();
						});			
						
					// If labels are enabled, add a text node
					if(chart.options.labels == true) {					
						d3.select(this).append('text')
							.attr("class", "chart-pie-label chart-pie-label-" + chart.options.labelPosition)
							.attr("pointer-events", "none");
					}
					
					// If selectable, add a selection arc
					if(chart.options.selectable == true) {
						d3.select(this).append('path')
							.style('fill', function(d, i) { return chart.options.color(d.data[chart.options.colorValue]); })
							.attr('class', 'chart-pie-select-slice selected');			
					}
				});
				
			// Update position of each arc
			pie.attr('transform', 'translate(' + chart.pie.x +  ',' + chart.pie.y + ')');
			
			// Timeout is only set if duration is greater than zero. This is to prevent issues
			//   with mouse hovers over pies that are being rendered.
			var timeout = chart.options.animation > 0 ? 200 : 0;
			
			// Render the pie chart. This is done in timeout for animated draws only.
			setTimeout(function(){
				// Render the pie
				pie.sort(function(a, b){ return BaseChart.sort(chart, a.data, b.data); })
					.select('.chart-pie-slice')
						.attr('class', function(d) {
							if(chart.scope.hover != undefined && chart.scope.hover[chart.options.uid] == d.data[chart.options.uid])
								return d3.select(this).attr('class') + ' hover';
							else
								return d3.select(this).attr('class');							
						})
						.transition()
						.ease("linear")
						.delay(function(d, i) { 
							// If this is a draw, then apply delay
							if(isAnimated && this._current.startAngle == d.startAngle && this._current.endAngle == d.endAngle)
								return i * (chart.options.animation);
							return 0;
						})
						.duration(chart.options.animation)
							.attrTween('d', function(d) {
								var arc = chart.pieArc;
								var isHover = chart.scope.hover != undefined && chart.scope.hover.d.data[chart.options.uid] == d.data[chart.options.uid];
								if(chart.options.growOnHover == true && isHover == true)
									arc = chart.growArc;
									
								// If this is a draw animate between 0 and new state
								if(isAnimated && this._current.startAngle == d.startAngle && this._current.endAngle == d.endAngle) {
									var i = d3.interpolate(d.startAngle, d.endAngle);
									return function(t) {
										d.endAngle = i(t);
										return arc(d);
									}
								}
								// Otherwise animate between old and new state
								else if(isAnimated) {
									var interpolate = d3.interpolate(this._current, d);
									this._current = interpolate(0);
									return function(t) {
										return arc(interpolate(t));
									};
								}
								else {
									return function(t) {
										return arc(d);
									};
								}
							})
						.each("end", function(d) {
							// If there's a tooltip, update the percentage
							if(chart.scope.tooltipData != undefined && chart.scope.tooltipData.payload.data[chart.options.uid] == d.data[chart.options.uid]) {
								var angle = (chart.pie.endAngle < 0 ? 360 + chart.pie.endAngle : chart.pie.endAngle) - chart.pie.startAngle;
								var pct = Math.abs((d.endAngle - d.startAngle) / (angle * Math.PI / 180));
								chart.scope.tooltipData.percentage = pct;
							}
							
							// If labels are enabled, update text and position
							if(chart.options.labels == true) {
								d3.select(this.parentNode).select('.chart-pie-label')
									.text(function() { 
										var angle = (chart.pie.endAngle < 0 ? 360 + chart.pie.endAngle : chart.pie.endAngle) - chart.pie.startAngle;
										var pct = Math.abs((d.endAngle - d.startAngle) / (angle * Math.PI / 180));
										var disp = d.data[chart.options.value];
										if(chart.options.labelType == "key")
											disp = d.data[chart.options.key];
										else if(chart.options.labelType == "percent")
											disp = pct;
										if(chart.options.labelFormat != undefined && chart.options.labelFormat != "")
											disp = eval("(" + chart.options.labelFormat + ")")(disp);
										else if(chart.options.labelType == "percent")
											disp = d3.format(".1%")(disp);									
										if(pct > chart.options.labelThreshold)
											return disp; 
									})
									.attr("transform", function() {
										var radius = chart.pie.outerRadius;
										if(chart.options.labelPosition == "outside")
											radius = chart.pie.labelRadius;
										var x = (radius - 22) * Math.sin(((d.endAngle - d.startAngle) / 2) + d.startAngle)  - 0.5 * this.getBBox().width;
										var y = (-1 * (radius - 20) * Math.cos(((d.endAngle - d.startAngle) / 2) + d.startAngle)) + 0.5 * this.getBBox().height;
										return "translate(" + x + "," + y + ")"; 
									});								
							}
						});
					
				// Render the pie select slice
				pie.select('.chart-pie-select-slice')
					.attr('d', function(d) { return chart.selectArc(d); })
					.style('opacity', function(d) { 
						if(chart.scope.selected != undefined && chart.scope.selected[chart.options.uid] == d.data[chart.options.uid])
							return null;
						return 0; 
					});
				
				// Enable hover actions
				setTimeout(function() {
					delete chart.hoverEnabled;
				}, chart.options.animation + 250);
			}, timeout);
		},
		
		calculateRadii: function(chart) {
			// Setup constants
			var selectOuterRadiusFactor = 10;
			var selectInnerRadiusFactor = 1;
			var growOnHoverRadiusFactor = 5;
			var labelRadiusFactor 		= chart.options.labelRadius;
			
			// Reset properties
			chart.pie.outerRadius		= 0;
			chart.pie.innerRadius		= 0;
			chart.pie.growOnHoverRadius	= 0;
			chart.pie.selectInnerRadius	= 0;
			chart.pie.selectOuterRadius	= 0;
			chart.pie.labelRadius		= 0;
			chart.pie.groupInnerRadius	= 0;
			chart.pie.groupOuterRadius	= 0;
			
			var plotArea = chart.plotArea.area;					
			var diam = Math.min(chart.plotArea.height, chart.plotArea.width);

			// Start with quadrants if enabled
			if(chart.options.quadrant != undefined) {
				// N, S, is half pie horizontal
				if(chart.options.quadrant == "N" || chart.options.quadrant == "S")
					chart.pie.outerRadius = Math.min(chart.plotArea.height, chart.plotArea.width / 2);
				// N, S, is half pie vertical
				else if(chart.options.quadrant == "E" || chart.options.quadrant == "W")
					chart.pie.outerRadius = Math.min(chart.plotArea.width, chart.plotArea.height / 2);					
				// NE, NW, SE, SW is quarter pie
				else if(chart.options.quadrant == "NE" || chart.options.quadrant == "NW" || chart.options.quadrant == "SE" || chart.options.quadrant == "SW")
					chart.pie.outerRadius = Math.min(chart.plotArea.width, chart.plotArea.height);					
			}

			// If not yet set, then set to default of available radius
			if(chart.pie.outerRadius == 0)
				chart.pie.outerRadius = diam / 2;

			// If outside labels are enabled
			if(chart.options.labels == true && chart.options.labelPosition == "outside") {
				chart.pie.labelRadius = chart.pie.outerRadius;
				chart.pie.outerRadius = chart.pie.labelRadius - labelRadiusFactor;
			}
			
			// If the pie is selectable or growOnHover
			if(chart.options.selectable == true || chart.options.growOnHover == true) {
				var excess = Math.max(selectOuterRadiusFactor, growOnHoverRadiusFactor);
				if(chart.pie.labelRadius == 0)
					chart.pie.outerRadius = chart.pie.outerRadius - excess;
				
				if(chart.options.selectable == true) {
					chart.pie.selectOuterRadius = chart.pie.outerRadius + selectOuterRadiusFactor;
					chart.pie.selectInnerRadius = chart.pie.outerRadius + selectInnerRadiusFactor;
				}
				if(chart.options.growOnHover == true)
					chart.pie.growOnHoverRadius = chart.pie.outerRadius + growOnHoverRadiusFactor;
			}
			
			// If the chart is a donut calculate the inner radius
			if(chart.options.donut == true) 
				chart.pie.innerRadius = chart.pie.outerRadius * chart.options.donutRatio;	

			// If the chart has a group by then calculate the group radius
			if(chart.options.groupBy == true) {
				chart.pie.groupInnerRadius = chart.pie.innerRadius;
				chart.pie.groupOuterRadius = (chart.pie.outerRadius - chart.pie.innerRadius) * 0.5 + chart.pie.innerRadius;			
				chart.pie.innerRadius = chart.pie.groupOuterRadius;
			}
		}
	};
});
