package com.streambase.fx.adminclient;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.SocketAddress;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/*
 * This class is the Main class of the Streambase SBTradingComponentsAdmin
 * It is responsible for a REPL on StdIn/StdOut to interact with the user
 * 
 * Most other classes in this package are instantiated or called directly by methods in this one
 */
public class SBTradingComponentsAdmin {
    static final String DELIMITER = "&";
    static long POLL_INTERVAL = 100;
    static long SOCKET_TIMEOUT = 5000;

    private static final int DEFAULT_PORT = 9000;
    private static final String DEFAULT_UDP_HOST = "239.0.0.0";
    private static final String START_USAGE = "\nsb-trading-components-admin [options]\n" + "Options: [-help] Displays this text; \n" + "\t ["
            + UIUtil.HOST_COMMAND + "] multicast_group_host [default:" + DEFAULT_UDP_HOST + "];\n" + "\t [" + UIUtil.PORT_COMMAND
            + "] multicast_port_number [default:" + DEFAULT_PORT + "];\n" + "\t [" + UIUtil.DEBUG_COMMAND + "] sets the debug flag to show more information;\n";
    private UDPListenerThread udpListenerThread;
    private TCPClientSelector tcpClientSelector;
    private Thread tcpClientSelectorThread;
    private BufferedReader stdIn = new BufferedReader(new InputStreamReader(System.in));
    private List<String> commandLineArguments;
    public static boolean debugMode = false;

    public SBTradingComponentsAdmin(String[] commandLineArguments) {
        this.commandLineArguments = new ArrayList<String>(Arrays.asList(commandLineArguments));
    }

    public void start() {
        if (commandLineArguments.indexOf("--help") >= 0 || commandLineArguments.indexOf("-help") >= 0 || commandLineArguments.indexOf("-h") >= 0 || commandLineArguments.indexOf("/?") >= 0
                || commandLineArguments.indexOf("-?") >= 0 || commandLineArguments.indexOf("--?") >= 0) {
            System.out.println("\nCommand Line Usage: " + START_USAGE);
            return;
        }
        if (commandLineArguments.indexOf(UIUtil.DEBUG_COMMAND) >= 0) {
            debugMode = true;
            commandLineArguments.remove(commandLineArguments.indexOf(UIUtil.DEBUG_COMMAND));
        }
        String multicastGroupString = DEFAULT_UDP_HOST;
        if (commandLineArguments.indexOf(UIUtil.HOST_COMMAND) >= 0) {
            int hostIndex = commandLineArguments.indexOf(UIUtil.HOST_COMMAND);
            multicastGroupString = commandLineArguments.size() > hostIndex + 1 ? commandLineArguments.get(hostIndex + 1) : DEFAULT_UDP_HOST;
        }
        int port = DEFAULT_PORT;
        if (commandLineArguments.indexOf(UIUtil.PORT_COMMAND) >= 0) {
            int portIndex = commandLineArguments.indexOf(UIUtil.PORT_COMMAND);
            port = commandLineArguments.size() > portIndex + 1 ? Integer.valueOf(commandLineArguments.get(portIndex + 1)).intValue() : DEFAULT_PORT;
        }

        InetAddress mcastGroup;
        try {
            mcastGroup = InetAddress.getByName(multicastGroupString);
            if (!mcastGroup.isMulticastAddress()) {
                System.err.println(multicastGroupString
                        + " is not/does not resolve to a valid multicast IP address.\nValid multicast IPs are in the range: 224.0.0.1 <= multicastIP <= 239.255.255.255");
                return;
            }
        } catch (UnknownHostException e) {
            System.err.println(multicastGroupString + " does not resolve to a valid IP address");
            return;
        }

        try {
            if (!UIUtil.isValidPortNum(port)) {
                System.err.println("Invalid port number: " + port);
                return;
            }
        } catch (Exception e) {
            System.err.println("Invalid port number: " + port);
            return;
        }

        System.out.println("Acquiring connections to all listening Streambase FX Servers on MulticastIP " + mcastGroup.toString() + " Port " + port);
        System.out.println("\nType 'help' for a list of commands and usage.\nType 'quit' to end the program.");
        start(port, mcastGroup);
    }

    private boolean createCoreObjects(int port, InetAddress mcastGroup) {
        try {
            if (tcpClientSelector == null) {
                tcpClientSelector = new TCPClientSelector();
                tcpClientSelectorThread = new Thread(tcpClientSelector, "TCPClient Selector");
                tcpClientSelectorThread.setUncaughtExceptionHandler(new Thread.UncaughtExceptionHandler() {
                    public void uncaughtException(Thread t, Throwable e) {
                        System.err.println("Unhandled error in client selector: " + e);

                    }
                });
                tcpClientSelectorThread.start();
            }
        } catch (IOException e) {
            System.err.println("Failed to create the client selector: " + e);
            return false;
        }
        // Establish all of our connections
        try {
            createUDPConnections(port, mcastGroup);
        } catch (Exception e) {
            System.err.println("Failed to create the UDP connection: " + e);
            return false;
        }
        return true;
    }

    private void start(int port, InetAddress mcastGroup) {
        try {
            if (!createCoreObjects(port, mcastGroup)) {
                return;
            }

            while (true) {
                String userInput = readInputDoLocalCmd(stdIn);
                // Special Commands
                if (userInput == null || userInput.equals(UIUtil.QUIT)) {
                    return;
                }
                if (userInput.equals(UIUtil.CONTINUE) || userInput.replaceAll(UIUtil.WHITESPACE, "").isEmpty()) {
                    continue;
                }
                if (userInput.equals(UIUtil.PING)) {
                    udpListenerThread.ping();
                    continue;
                }

                try {
                    tcpClientSelector.writeToClient(UIUtil.generateCommandData(userInput));
                } catch (Exception e) {
                    System.err.println("ERROR: " + e.getMessage());
                    continue;
                }
            }
        } finally {
            cleanUp();
        }
    }

    private void cleanUp() {
        // Make sure our UDPListener is stopped
        if (udpListenerThread != null) {
            udpListenerThread.interrupt();
        }
        if (tcpClientSelectorThread != null) {
            tcpClientSelector.stop();
            try {
                tcpClientSelectorThread.join();
            } catch (InterruptedException e) {
                System.err.println("Failed to close the client selector thread: " + e);
            }
        }
        try {
            stdIn.close();
        } catch (IOException e) {
        } // just ignore and go
    }

    // Spawn a new thread, which will send out UDP requests and then listen for responses that can add to this objects
    // list of connections via the static variable active_sockets
    private void createUDPConnections(int port, InetAddress mcastGroup) throws IOException {
        SocketAddress addr = getUDPSocketAddress();
        DatagramSocket listenSock = new DatagramSocket(null);
        listenSock.setReuseAddress(true);
        listenSock.bind(addr);
        udpListenerThread = new UDPListenerThread(listenSock, port, mcastGroup, tcpClientSelector);
        udpListenerThread.start();
    }

    // this creates a dummy socket just to get a local free socket address to use for UDP listen
    private SocketAddress getUDPSocketAddress() throws SocketException {
        DatagramSocket udpSock = new DatagramSocket();
        udpSock.setReuseAddress(true);
        SocketAddress addr = udpSock.getLocalSocketAddress();
        udpSock.close();
        return addr;
    }

    /*
     * If userInput equals null, will terminate program If userInput equals QUIT it will verify this command If
     * userInput equals LIST_CONNECTIONS it will return that string; If userInput equals PING it will cause the listener
     * thread to ping the multicast group If userInput equals HELP it will print a help messasge This allows us to
     * execute all local commands before command verification Else return the exact user input
     */
    public String readInputDoLocalCmd(BufferedReader stdIn) {
        String userInput = UIUtil.CONTINUE;
        try {
            // Read User Input
            System.out.println("Ready for command:");
            userInput = stdIn.readLine();
            if (userInput != null) {
                userInput = userInput.trim();
            }
            if (userInput == null || userInput.equals(UIUtil.QUIT) || userInput.equals(UIUtil.SHORT_QUIT)) {
                return handleQuit(stdIn, userInput);
            } else if (userInput.equals(UIUtil.LIST_CONNECTIONS) || userInput.equals(UIUtil.SHORT_LIST_CONNECTIONS)) {
                tcpClientSelector.printActiveConnections();
                return UIUtil.CONTINUE;
            } else if (userInput.equals(UIUtil.HELP) || userInput.equals(UIUtil.SHORT_HELP)) {
                UIUtil.printHelpInfo();
                return UIUtil.CONTINUE;
            } else if (userInput.equals(UIUtil.PING) || userInput.equals(UIUtil.SHORT_PING)) {
                return UIUtil.PING;
            } else if (userInput.startsWith(UIUtil.HELP) || userInput.startsWith(UIUtil.SHORT_HELP)) {
                String[] help = userInput.split(UIUtil.WHITESPACE);
                if (help.length >= 2 && (help[0].equals(UIUtil.HELP) || help[0].equals(UIUtil.SHORT_HELP))) {
                    UIUtil.printHelpInfo(help[1]);
                } else {
                    return userInput;
                }
                return UIUtil.CONTINUE;
            } else {
                return userInput;
            }
        } catch (IOException e) {
            e.printStackTrace();
            return UIUtil.QUIT;
        }
    }

    private String handleQuit(BufferedReader stdIn, String userInput) throws IOException {
        if (userInput == null) {
            System.out.println("Termination Signaled,\nExiting Admin Client now.");
            return UIUtil.QUIT;
        }
        String quitMessage = "Are you sure you would like to quit?\n" + "All data not written to a file will be lost. [y\\n]\n";
        if (userInput.equals(UIUtil.QUIT) || userInput.equals(UIUtil.SHORT_QUIT)) {
            System.out.print(quitMessage);
            userInput = stdIn.readLine();
            while (userInput != null && !userInput.equals("y") && !userInput.equals("n") && !userInput.equals("yes") && !userInput.equals("no")) {
                System.out.print(quitMessage);
                userInput = stdIn.readLine();
            }
        } else {
            return UIUtil.QUIT;
        }
        if (userInput.equals("y") || userInput.equals("yes")) {
            System.out.println("Termination Signaled,\nExiting Admin Client now.");
            return UIUtil.QUIT;
        } else {
            return UIUtil.CONTINUE;
        }
    }

    // This will be called upon command line invocation of the jar
    public static void main(String[] args) {
        new SBTradingComponentsAdmin(args).start();
    }
}
