package com.streambase.fx.adminclient;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.SocketChannel;
import java.util.HashMap;
import java.util.Map;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.streambase.sb.util.ByteOrderedDataOutput;

/*
 * This is the TCP client part which directly talks to the FX framework.
 * This class holds the connection information and deals with reading and writing to the TCP socket (channel).
 */
public class TCPClient {
    public static final String HTTP_HEADER_SEPARATOR = "\r\n\r\n";
    public static final String HTTP_CONTENT_LENGTH = "Content-Length";
    private SocketChannel channel;
    private ByteBuffer readBuffer = ByteBuffer.allocate(1024 * 10);
    private ByteBuffer writeBuffer = ByteBuffer.allocate(1024 * 10);
    private Map<String, TCPClientRequest> tcpClientRequests = new HashMap<String, TCPClientRequest>();
    private String host = "";
    private int port = 0;
    private String venueName = "";
    private TCPClientSelector tcpClientSelector;

    public TCPClient(String host, int port, String venueName, TCPClientSelector tcpClientSelector) {
        this.host = host;
        this.port = port;
        this.venueName = venueName;
        this.tcpClientSelector = tcpClientSelector;
        writeBuffer.clear();
        readBuffer.clear();
    }

    static public String createKey(String host, int port) {
        return host + ":" + port;
    }
    
    public String getKey() {
        return TCPClient.createKey(host, port);
    }

    public String getFileKey() {
        return host + "_" + port;
    }

    public String getDisplayName() {
        return venueName + " [" + getKey() + "]";
    }

    public String getLocalHostAddress() {
        return channel.socket().getLocalAddress().getHostAddress();
    }

    public int getLocalPort() {
        return channel.socket().getLocalPort();
    }

    public void setChannel(SocketChannel channel) {
        this.channel = channel;
    }

    public Boolean read() {
        Boolean readPerformed = false;
        try {
            int readCount = channel.read(readBuffer);
            if (readCount < 0) {
                close(true);
                return null;
            }
            if (readCount > 0) {
                readPerformed = true;
                readBuffer.flip(); // flip the buffer to allow for read operations
                while (readBuffer.hasRemaining()) {
                    try {
                        String httpData = new String(readBuffer.array(), readBuffer.position(), readBuffer.limit());
                        int position = 0;
                        int headerEndPos = httpData.indexOf(HTTP_HEADER_SEPARATOR, position);
                        if (headerEndPos > position) {
                            String headerData = httpData.substring(position, headerEndPos);
                            position = headerEndPos + HTTP_HEADER_SEPARATOR.length();
                            readBuffer.position(readBuffer.position() + position); // set position just in case body fails
                            String[] headerSplit = headerData.split("\n");
                            int contentLength = 0;
                            for (String headerSection : headerSplit) {
                                String [] keyvalue = headerSection.split(":");
                                if (keyvalue.length == 2 && keyvalue[0].toLowerCase().equals(HTTP_CONTENT_LENGTH.toLowerCase())) {
                                    contentLength = keyvalue[1].trim().length() > 0 ? Integer.parseInt(keyvalue[1].trim()) : 0;        
                                }
                            }
                            int remainingData = (httpData.length() - position);
                            if ((contentLength > 0 || remainingData == 0) && remainingData >= contentLength) {
                                String contentData = httpData.substring(position, position + contentLength);
                                readBuffer.position(readBuffer.position() + contentLength);
                                Object jsonResponse = JSON.parse(contentData);
                                if (jsonResponse instanceof JSONObject) {
                                    if (SBTradingComponentsAdmin.debugMode) {
                                        System.out.println(String.format("Received JSON from %s:\n%s", getDisplayName(), contentData));
                                    }
                                    parseJSONObject((JSONObject) jsonResponse, contentData);
                                }
                            }
                        }
                    } catch (Exception e) {
                        System.err.println(String.format("Error reading data from client %s: %s", getDisplayName(), e));
                        break;
                    }
                }
                readBuffer.compact();
            }
        } catch (ClosedChannelException e) {
            close();
            return null;
        } catch (IOException e) {
            close();
            return null;
        }
        return readPerformed;
    }

    private void parseJSONObject(JSONObject jsonResponse, String dataReceived) {
        if (jsonResponse.containsKey(TCPClientRequest.JSON_REQUESTID)) {
            String requestId = (String) jsonResponse.get(TCPClientRequest.JSON_REQUESTID);
            TCPClientRequest tcpClientRequest = null;
            synchronized (tcpClientRequests) {
                if (tcpClientRequests.containsKey(requestId)) {
                    tcpClientRequest = tcpClientRequests.get(requestId);
                    tcpClientRequest.setJSONResponse(jsonResponse);
                    tcpClientRequests.remove(requestId);
                }
            }
            if (tcpClientRequest == null) {
                System.out.println(String.format("Client %s request that could not be found:\n%s", getDisplayName(), requestId));
            } else {
                UIUtil.printTCPClientResponse(this, tcpClientRequest);
            }
        } else {
            System.out.println(String.format("Client %s returned invalid data:\n%s", getDisplayName(), dataReceived));
        }
    }

    public void write(TCPClientRequest tcpClientRequest) {
        synchronized (tcpClientRequests) {
            tcpClientRequests.put(tcpClientRequest.getRequestId(), tcpClientRequest);
        }
        synchronized (writeBuffer) {
            String jsonData = tcpClientRequest.getJSONObject().toJSONString();
            ByteOrderedDataOutput output = new ByteOrderedDataOutput(writeBuffer.limit(), ByteOrder.nativeOrder());
            String outString = "GET http://fxadmintool/ HTTP/1.1";
            outString += "\n" + HTTP_CONTENT_LENGTH + ": " + jsonData.length();
            outString += HTTP_HEADER_SEPARATOR + jsonData; // format as http request
            if (SBTradingComponentsAdmin.debugMode) {
                System.out.println(String.format("Sending JSON to %s:\n%s", getDisplayName(), outString));
            }
            output.writeBytes(outString);
            int size = output.getPosition();
            writeBuffer.put(output.getBuffer(), 0, size);
            try {
                output.close();
            } catch (IOException e) {
                System.err.println(String.format("Error closing client byte ordered writer %s: %s", getDisplayName(), e));
            }
            if (writeBuffer.hasRemaining()) {
                writeBuffer();
            }
        }
    }

    private void writeBuffer() {
        writeBuffer.flip();
        try {
            channel.write(writeBuffer);
        } catch (IOException e) {
            System.err.println(String.format("Error reading/writing from/to connection %s: %s", getDisplayName(), e));
            close();
        }
        writeBuffer.compact();
    }

    public void close() {
        close(false);
    }
    
    public void close(boolean reconnection) {
        if (reconnection) {
            tcpClientSelector.reconnectClient(host, port, venueName);
        } else {
            tcpClientSelector.removeClient(this);
            try {
                channel.close();
                System.out.println("Closing socket to: " + getDisplayName());                
            } catch (IOException e) {
                System.err.println(String.format("Error closing socket %s: %s", getDisplayName(), e));
            }
            tcpClientSelector.printActiveConnections();
        }
    }

}
