package com.streambase.fx.adminclient;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.SocketChannel;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Queue;
import java.util.Set;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.RejectedExecutionException;

/*
 * This is the TCP selector (nio channels) which handles checking if data is ready to read from a socket.
 */
public class TCPClientSelector implements Runnable {
    protected Selector selector;
    private boolean running = true;
    private final Map<String, TCPClient> socketChannels = Collections.synchronizedMap(new HashMap<String, TCPClient>());

    public TCPClientSelector() throws IOException {
        selector = Selector.open();
    }

    @Override
    public void run() {
        try {
            while (running && selector.isOpen()) {
                if (selector.select(100) <= 0) {
                    registerClientsFromQueue();
                    continue;
                }
                Set<?> readyKeys = selector.selectedKeys();
                Iterator<?> iterator = readyKeys.iterator();
                while (iterator.hasNext()) {
                    SelectionKey key = (SelectionKey) iterator.next();
                    iterator.remove();
                    if (key.isValid()) {

                        TCPClient tcpClient = (TCPClient) key.attachment();
                        if (tcpClient != null) {
                            if (key.isReadable()) {
                                if (tcpClient.read() == null) {
                                    key.cancel();
                                }
                            }
                        }
                    }
                    if (!key.isValid()) {
                        key.cancel();
                    }
                }
            }
        } catch (RejectedExecutionException e) {
            if (running) {
                System.err.println("Client selector failed: " + e);
            }
        } catch (IOException e) {
            System.err.println("Error setting up client selector: " + e);
        } catch (Exception e) {
            System.err.println("Error in the client selector: " + e);
        }
    }

    private class RegisterObject {
        public TCPClient tcpClient;
        public SocketChannel channel;

        RegisterObject(TCPClient tcpClient, SocketChannel channel) {
            this.tcpClient = tcpClient;
            this.channel = channel;
        }
    }

    private Queue<RegisterObject> queue = new ConcurrentLinkedQueue<RegisterObject>();

    private void registerClientsFromQueue() {
        RegisterObject registerObject = null;
        while ((registerObject = queue.poll()) != null) {
            try {
                SelectionKey clientKey = registerObject.channel.register(selector, SelectionKey.OP_READ);
                clientKey.attach(registerObject.tcpClient);
            } catch (ClosedChannelException e) {
                System.err.println("Client socket selector error trying to register client channel: " + e);
            }
        }
    }

    public void RegisterClient(TCPClient tcpClient, SocketChannel channel) {
        synchronized (queue) {
            queue.add(new RegisterObject(tcpClient, channel));
        }
    }

    public void stop() {
        running = false;
        HashMap<String, TCPClient> connections = new HashMap<String, TCPClient>(socketChannels);
        Iterator<Entry<String, TCPClient>> it = connections.entrySet().iterator();
        while (it.hasNext()) {
            it.next().getValue().close();
        }
    }

    public void removeClient(TCPClient client) {
        synchronized (socketChannels) {
            socketChannels.remove(client.getKey());
        }
    }

    public void createClient(final String host, final int port, final String venueName) {        
        final String key = TCPClient.createKey(host, port);
        synchronized (socketChannels) {
            if (socketChannels.containsKey(key)) {
                return;
            }
        }
        createClientConnection(host, port, venueName, false);
    }
    
    public void reconnectClient(final String host, final int port, final String venueName) {
        createClientConnection(host, port, venueName, true);
    }
    
    private void createClientConnection(final String host, final int port, final String venueName, final boolean reconnect) {
        final TCPClient client = new TCPClient(host, port, venueName, this);
        Thread connectThread = new Thread(new Runnable() {
            @Override
            public void run() {
                InetSocketAddress addr = new InetSocketAddress(host, port);
                try {
                    synchronized (socketChannels) {
                        SocketChannel channel = SocketChannel.open(addr);
                        channel.configureBlocking(false);
                        client.setChannel(channel);
                        socketChannels.put(client.getKey(), client);
                        RegisterClient(client, channel);
                    }
                    if (!reconnect) {
                    	System.out.println(String.format("Connected to %s", client.getDisplayName()));
                    }
                    return;
                } catch (IOException e) {
                    System.err.println(String.format("Failed to connect to %s", client.getDisplayName()));
                } catch (Exception e) {
                    System.err.println(String.format("Failed to setup a connection to %s", client.getDisplayName()));
                }
            }
        });
        connectThread.setUncaughtExceptionHandler(new Thread.UncaughtExceptionHandler() {
            public void uncaughtException(Thread t, Throwable e) {
                System.err.println(String.format("Unhandled error in client %s connection thread: %s", client.getKey(), e));
            }
        });
        connectThread.start();
    }

    public void writeToClient(TCPClientRequest tcpClientRequest) {
        String connectionKey = tcpClientRequest.getSendToTCPClientKey();
        int indexRef = tcpClientRequest.getSendToTCPClientIndex();
        synchronized (socketChannels) {
            if (connectionKey.length() > 1) {
                if (socketChannels.containsKey(connectionKey)) {
                    writeToClient(socketChannels.get(connectionKey), tcpClientRequest);
                } else {
                    System.out.println(String.format("Connection %s not found", connectionKey));
                }
            } else {
                Iterator<Entry<String, TCPClient>> it = socketChannels.entrySet().iterator();
                int position = 0;
                boolean connectionFound = false;
                while (it.hasNext()) {
                    Entry<String, TCPClient> pairs = it.next();
                    if (indexRef < 0 || indexRef == position++) {
                        writeToClient(pairs.getValue(), tcpClientRequest);
                        connectionFound = true;
                        if (indexRef >= 0) {
                            break;
                        }
                    }
                }
                if (!connectionFound) {
                    if (indexRef < 0) {
                        System.out.println(String.format("No data was sent, please check you are connected to at least one server"));
                    } else {
                        System.out.println(String.format("No data was sent to index connection %s, please check you are connected to at least one server", indexRef));
                    }
                }
            }
        }
    }

    private void writeToClient(TCPClient client, TCPClientRequest tcpClientRequest) {
        client.write(tcpClientRequest);
    }

    public void printActiveConnections() {
        if (!running) {
            return;
        }
        System.out.println("Printing all current connections:");
        System.out.println("---------------------------------");
        synchronized (socketChannels) {
            if (socketChannels.isEmpty()) {
                System.out.println("There are no current connections.");
            } else {
                Iterator<Entry<String, TCPClient>> it = socketChannels.entrySet().iterator();
                int position = 0;
                while (it.hasNext()) {
                    Entry<String, TCPClient> pairs = it.next();
                    String out = "[" + position++ + "] : " + pairs.getValue().getDisplayName();
                    System.out.println(out);
                }
            }
        }
        System.out.println("---------------------------------");
    }
}
