package com.streambase.fx.adminclient;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.SocketException;
import java.net.SocketTimeoutException;

/*
 * This class Is instantiated and started by the SBTradingComponentsAdmin class, it is also interrupted
 * by that class to close its UDP socket at shutdown
 * 
 * This class is responsible for periodically 'pinging' a multicast group, listening for
 * replies, and then connecting to the sockets specified in those replies
 */
public class UDPListenerThread extends Thread {
    public final static int DEFAULT_SOCKET_TIMEOUT = 5000;
    private int socketTimeout;
    private DatagramSocket sock;
    private int port;
    private InetAddress mcastGroup;
    private final String udpPingMessage = "PING";
    private TCPClientSelector tcpClientSelector;

    public UDPListenerThread(DatagramSocket sock, int port, InetAddress mcastGroup, TCPClientSelector tcpClientSelector) {
        this.sock = sock;
        this.port = port;
        this.mcastGroup = mcastGroup;
        this.socketTimeout = DEFAULT_SOCKET_TIMEOUT;
        this.tcpClientSelector = tcpClientSelector;
    }

    @Override
    public void run() {
        try {
            sock.setSoTimeout(socketTimeout);
        } catch (SocketException e) {
            System.err.println("Error setting UDP socket timeout: " + e);
        }
        ping();
        try {
            while (true) {
                // Listen for incoming UDP packets
                byte[] buff = new byte[1024];
                DatagramPacket in = new DatagramPacket(buff, buff.length);
                try {
                    sock.receive(in);
                } catch (SocketTimeoutException e) {
                    ping();
                    continue;
                }
                try {
                    String message = new String(in.getData(), 0, in.getLength());
                    String[] values = message.split(",");
                    tcpClientSelector.createClient(in.getAddress().getHostAddress(), Integer.valueOf(values[0]).intValue(), values[1]);
                } catch (Exception e) {
                    if (Thread.interrupted()) {
                        throw e;
                    } else {
                        System.err.println("UDP Message failed: " + e);
                    }
                }
            }
        } catch (Exception e) {
            if (!Thread.interrupted()) {
                System.err.println("UDP listener failed: " + e);
            }
        } finally {
            sock.close();
        }
    }

    // Send out a notify message on the Multicast Group, to acquire new connections
    public void ping() {
        DatagramPacket out = new DatagramPacket(udpPingMessage.getBytes(), udpPingMessage.length(), mcastGroup, port);
        try {
            sock.send(out);
        } catch (IOException e) {
            System.err.println("Connection to multicast group lost.\nUse the 'quit' command to stop the program.");
        }
    }

    @Override
    public void interrupt() {
        super.interrupt();
        this.sock.close();
    }
}
